import polars as pl
import pytest
from polars.testing import assert_frame_equal

from maplib import BlankNode

pl.Config.set_fmt_str_lengths(300)


@pytest.fixture(scope="function")
def pizzas_mapping():
    from maplib import (
        Mapping,
        Prefix,
        Template,
        Argument,
        Parameter,
        Variable,
        RDFType,
        Triple,
        a,
    )
    import polars as pl

    pl.Config.set_fmt_str_lengths(150)

    pi = "https://github.com/DataTreehouse/maplib/pizza#"
    df = pl.DataFrame(
        {
            "p": [pi + "Hawaiian", pi + "Grandiosa"],
            "c": [pi + "CAN", pi + "NOR"],
            "ings": [[pi + "Pineapple", pi + "Ham"], [pi + "Pepper", pi + "Meat"]],
        }
    )
    # print(df)

    pi = Prefix("pi", pi)

    p_var = Variable("p")
    c_var = Variable("c")
    ings_var = Variable("ings")

    template = Template(
        iri=pi.suf("PizzaTemplate"),
        parameters=[
            Parameter(variable=p_var, rdf_type=RDFType.IRI()),
            Parameter(variable=c_var, rdf_type=RDFType.IRI()),
            Parameter(variable=ings_var, rdf_type=RDFType.Nested(RDFType.IRI())),
        ],
        instances=[
            Triple(p_var, a(), pi.suf("Pizza")),
            Triple(p_var, pi.suf("fromCountry"), c_var),
            Triple(p_var, pi.suf("hasBlank"), BlankNode("MyBlank")),
            Triple(
                p_var,
                pi.suf("hasIngredient"),
                Argument(term=ings_var, list_expand=True),
                list_expander="cross",
            ),
        ],
    )

    m = Mapping()
    m.expand(template, df)
    hpizzas = """
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>
    CONSTRUCT { ?p a pi:HeterodoxPizza } 
    WHERE {
        ?p a pi:Pizza .
        ?p pi:hasIngredient pi:Pineapple .
    }"""
    m.insert(hpizzas)
    return m


def test_simple_query_no_error(pizzas_mapping):
    res = pizzas_mapping.query(
        """
    PREFIX pi:<https://github.com/DataTreehouse/maplib/pizza#>

    SELECT ?p WHERE {
    ?p a pi:HeterodoxPizza
    }
    """
    )

    expected_df = pl.DataFrame(
        {"p": ["<https://github.com/DataTreehouse/maplib/pizza#Hawaiian>"]}
    )
    assert_frame_equal(res, expected_df)



from maplib import Mapping, Template, IRI, Triple, Variable


def add_triples(source: Mapping, target: Mapping, source_graph: str = None, target_graph: str = None):
    for p in source.get_predicate_iris(source_graph):
        subject = Variable("subject")
        object = Variable("object")
        template = Template(
            iri=IRI("urn:maplib:tmp"),
            parameters=[subject, object],
            instances=[Triple(subject, p, object)]
        )
        sms = source.get_predicate(p, source_graph)
        for sm in sms:
            target.expand(
                template,
                sm.mappings,
                unique_subset=["subject", "object"],
                types=sm.rdf_types,
                graph=target_graph
            )



import polars as pl
from maplib import Mapping, RDFType, XSD
from polars.testing import assert_frame_equal
import polars as pl
pl.Config.set_fmt_str_lengths(200)

def test_multi_filter_equals():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(?a = :hello2)
    }
    """
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))

def test_multi_filter_numerical_equals():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
    {VALUES (?a) { ("1"^^xsd:int) }}
    {VALUES (?b) { ("1"^^xsd:integer) }}
    FILTER(?a = ?b)
    }
    """
    )
    assert df.shape == (0,2)

def test_multi_filter_numerical_cast_real_equals():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
    {VALUES (?a) { ("1"^^xsd:int) }}
    {VALUES (?b) { ("1"^^xsd:integer) }}
    FILTER(xsd:double(?a) = xsd:double(?b))
    }
    """
    )
    assert df.shape == (1,2)

def test_multi_filter_numerical_cast_integer_equals():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
        {VALUES (?a) { ("1"^^xsd:int) }}
        {VALUES (?b) { ("1"^^xsd:integer) }}
    FILTER(xsd:integer(?a) = xsd:integer(?b))
    }
    """
    )
    assert df.shape == (1,2)

def test_multi_filter_numerical_cast_string_equals():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b WHERE {
    VALUES (?a) { ("1"^^xsd:int)}
    VALUES (?b) { ("1"^^xsd:integer)} 
    FILTER(xsd:string(?a) = xsd:string(?b))
    }
    """
    )
    assert df.shape == (1,2)

def test_multi_filter_uri_cast_string_equals():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a  WHERE {
        { VALUES (?a) { ("1"^^xsd:int) }} UNION { VALUES (?a) { (:something) }}
        FILTER(xsd:string(?a) = "http://example.net/something")
    }
    """
    )
    assert df.shape == (1,1)

def test_filter_uri_cast_string_equals():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a  WHERE {
        VALUES (?a) { (:something) }
        FILTER(xsd:string(?a) = "http://example.net/something")
    }
    """
    )
    assert df.shape == (1,1)

def test_multi_filter_equals_with_datatypes():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(?a = :hello2)
    }
    """,
        include_datatypes=True,
    )
    assert sm.rdf_types == {"a": RDFType.IRI()}
    assert_frame_equal(
        sm.mappings, pl.DataFrame({"a": ["<http://example.net/hello2>"]})
    )

def test_multi_value_different_types_filter_equals_with_datatypes():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) ("string") (1) (true) }}
    FILTER(?a = true)
    }
    """,
        include_datatypes=True,
    )
    assert sm.rdf_types == {"a": RDFType.Literal(XSD().boolean)}
    assert_frame_equal(
        sm.mappings, pl.DataFrame({"a": [True]})
    )

def test_multi_value_different_types_filter_isiri_or_equals_with_datatypes():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) ("string") (1) (true) }}
    FILTER(isIri(?a) || ?a = 1)
    }
    """,
        include_datatypes=True,
    )
    assert sm.rdf_types == {"a": RDFType.Multi([RDFType.IRI(), RDFType.Literal(XSD().integer)])}
    assert_frame_equal(
        sm.mappings, pl.DataFrame({"a": ["<http://example.net/hello2>", '"1"^^<http://www.w3.org/2001/XMLSchema#integer>']})
    )

def test_coalesce_both_multi_same_types():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    {VALUES (?a) { (:hello1) ("string1") (1) (true) }}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2") (1) (false) }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0

def test_coalesce_both_multi_different_types():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    {VALUES (?a) { (:hello1) ("string1") }}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2"@en) (1) (false) }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0

def test_coalesce_both_multi_different_types_rhs_less():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) ("string1") (1)}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2") }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0

def test_coalesce_only_lhs_multi():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) ("string1") (1)}
    OPTIONAL {
        VALUES (?b) { (:hello2) (:hello3) }
        FILTER(isIri(?a))
    }
    }
    """,
        include_datatypes=True,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0

def test_coalesce_only_rhs_multi():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) (:hello3) (:hello4)}
    OPTIONAL {
        VALUES (?b) { (:hello2) ("string2"@no) ("string3") }
        FILTER(?a = :hello1)
    }
    }
    """,
        include_datatypes=True,
    )
    assert sm.mappings.get_column("c").is_null().sum() == 0


def test_coalesce_no_multi():
    m = Mapping([])
    sm = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a ?b (COALESCE(?b, ?a) AS ?c) WHERE {
    VALUES (?a) { (:hello1) (:hello2) (:hello3)}
    OPTIONAL {
        VALUES (?b) { (:hello5) (:hello6) }
        FILTER(?a = :hello1)
    }
    }
    """,
        include_datatypes=True,
    )
    assert sm.rdf_types == {'c': RDFType.IRI(), 'b': RDFType.IRI(), 'a': RDFType.IRI()}
    assert sm.mappings.get_column("c").is_null().sum() == 0

def test_multi_filter_equals_mirrored():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(:hello2 = ?a)
    }
    """
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))


def test_multi_filter_equals_two_multi():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    {VALUES (?b) { (:hello2) }} UNION {VALUES (?b) { (1) }}
    FILTER(?b = ?a)
    }
    """
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))


def test_multi_filter_is_in():
    m = Mapping([])
    df = m.query(
        """
    PREFIX : <http://example.net/> 
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT ?a WHERE {
    {VALUES (?a) { (:hello2) }} UNION {VALUES (?a) { ("string") }}
    FILTER(?a in (:hello2, "a"))
    }
    """
    )
    assert_frame_equal(df, pl.DataFrame({"a": ["<http://example.net/hello2>"]}))


import pathlib

from .maplib import *
from .add_triples import add_triples
PATH_HERE = pathlib.Path(__file__).parent.resolve()

if (PATH_HERE / "graph_explorer").exists():
    from .graph_explorer import explore


import pytest
from maplib import Mapping, Triple, Variable, Template, IRI, Parameter, RDFType
import polars as pl

def test_many_validation_errors_subject():
    n = 10_000
    df = pl.DataFrame({
        "subject": [f"MyBadIRI//{i}" for i in range(n)],
        "predicate": ["https://example.net/hasObject"]*n,
        "object": [i for i in range(n)]
    })
    m = Mapping()
    subj = Variable("subject")
    pred = Variable("predicate")
    obj = Variable("object")
    t = Template(
        IRI("https://github.com/DataTreehouse/maplib/my_template"),
        [subj, pred, obj],
        [Triple(subj, pred, obj)]
                 )
    with pytest.raises(Exception) as e:
        m.expand(t, df, validate_iris=True)
    assert "invalid IRI" in str(e)

def test_many_validation_errors_predicate():
    n = 10_000
    df = pl.DataFrame({
        "subject": [f"urn:dth:goodiri_{i}" for i in range(n)],
        "predicate": ["!https://example.net/hasObject"]*n,
        "object": [i for i in range(n)]
    })
    m = Mapping()
    subj = Variable("subject")
    pred = Variable("predicate")
    obj = Variable("object")
    t = Template(
        IRI("https://github.com/DataTreehouse/maplib/my_template"),
        [subj, pred, obj],
        [Triple(subj, pred, obj)]
    )
    with pytest.raises(Exception) as e:
        m.expand(t, df, validate_iris=True)
    assert "invalid IRI" in str(e)

def test_many_validation_errors_nested_object():
    n = 1_000
    df = pl.DataFrame({
        "subject": [f"urn:dth:goodiri_{i}" for i in range(n)],
        "predicate": ["https://example.net/hasObject"]*n,
        "object": [[f"aTerribleIRI!!{i}", f"anotherTerribleIRI!!{i}"] for i in range(n)]
    })
    m = Mapping()
    subj = Variable("subject")
    pred = Variable("predicate")
    obj = Variable("object")
    t = Template(
        IRI("https://github.com/DataTreehouse/maplib/my_template"),
        [subj, pred, Parameter(obj, rdf_type=RDFType.Nested(RDFType.IRI()))],
        [Triple(subj, pred, obj)]
    )
    with pytest.raises(Exception) as e:
        m.expand(t, df, validate_iris=True)
    assert "invalid IRI" in str(e)